import sys
from tempfile import TemporaryFile
from io import BytesIO, DEFAULT_BUFFER_SIZE
# > Typing
from types import TracebackType
from typing import Optional, IO, Callable, Literal, List
# > Local Imports
from .crc import CRC32
from .chunk import Chunk
from .types import CFMode, FPType
from .units import DEFAULT_CHUNK_FILE_SIGNATURE
from .functional import initfp, is_cfmode, iocopy, hasitem, formater
from .exceptions import IsNotChunkFileError, IONotWritableError, ChunkIsDamagedError

# ! Main Class Functions
def otempfile():
    return TemporaryFile("wb+", suffix="chunk")

def obytesio():
    return BytesIO()

# ! Main Class
class ChunkFile:
    # ! Private Initialization Methods
    def __ischunkfile(self) -> bool:
        self.__io.seek(0)
        return self.__io.read(len(self.__cfs)) == self.__cfs
    
    def __initchunkfile(self) -> None:
        if not self.__ischunkfile():
            return IsNotChunkFileError()
        self.__io.seek(len(self.__cfs))
        size_data = self.__io.read(self.__cls)
        while len(size_data) > 0:
            size = int.from_bytes(size_data, self.__bord)
            name = self.__io.read(self.__cns).decode(self.__encoding, self.__errors)
            cio = self.__openio()
            currect_crc = iocopy(self.__io, cio, size, self.__bfs, self.__bord)
            file_crc = CRC32.from_bytes(self.__io.read(4), self.__bord)
            if (int(file_crc) != int(currect_crc)) and (not self.__icrc):
                raise ChunkIsDamagedError()
            self.__chunks.append(Chunk(name, cio, self.__mode, chunk_name_size=self.__cns))
            size_data = self.__io.read(self.__cls)
    
    # ! Initialization
    def __init__(
        self,
        fp: FPType,
        mode: CFMode='r',
        *,
        buffer_size: int=DEFAULT_BUFFER_SIZE,
        chunk_file_signature: bytes=DEFAULT_CHUNK_FILE_SIGNATURE,
        chunk_name_size: int=4,
        chunk_length_size: int=4,
        openio: Callable[[], IO[bytes]]=otempfile,
        encoding: str='utf-8',
        errors: str='strict',
        byteorder: Literal['little', 'big']=sys.byteorder,
        ignore_crc: bool=False
    ) -> None:
        assert is_cfmode(mode)
        self.__name, self.__io = initfp(fp, mode)
        self.__mode = mode
        self.__chunks: List[Chunk] = []
        self.__bfs: int = buffer_size
        self.__cfs: bytes = chunk_file_signature
        self.__cns: int = chunk_name_size
        self.__cls: int = chunk_length_size
        self.__openio = openio
        self.__encoding = encoding
        self.__errors = errors
        self.__bord = byteorder
        self.__icrc: bool = ignore_crc
        self.__initchunkfile()
    
    # ! ChunkFile Private Methods
    def __GetChunk(self, name: str) -> Optional[Chunk]:
        for chunk in self.__chunks:
            if chunk.name == name:
                return chunk
    
    # ! Magic Methods
    def __str__(self) -> str:
        return formater(name=self.__name, mode=self.__mode)
    
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__str__()})"
    
    def __getitem__(self, key: str):
        chunk = self.__GetChunk(key)
        if chunk is None:
            if not self.writable():
                raise IONotWritableError()
            chunk = Chunk(key, self.__openio(), self.__mode)
            self.__chunks.append(chunk)
        return chunk
    
    def __enter__(self):
        return self
    
    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType]
    ) -> None:
        self.close()
    
    # ! Private IO Methods
    def __ioflush(self) -> None:
        self.__ioclear()
        self.__io.seek(0)
        self.__io.write(self.__cfs)
        for chunk in self.__chunks:
            if not chunk.closed:
                self.__io.write(chunk.size.to_bytes(self.__cls, self.__bord))
                self.__io.write(chunk.name.encode(self.__encoding, self.__errors))
                chunk.seek(0)
                crc = iocopy(chunk._io, self.__io, chunk.size, self.__bfs, self.__bord)
                self.__io.write(bytes(crc))
    
    def __ioclear(self) -> None:
        self.__io.seek(0)
        self.__io.truncate(0)
    
    # ! Propertyes
    @property
    def name(self) -> Optional[str]:
        return self.__name
    
    @property
    def mode(self) -> CFMode:
        return self.__mode
    
    @property
    def chunks(self) -> List[Chunk]:
        return self.__chunks
    
    # ! ChunkFile Methods
    def chunk(self, name: str) -> Chunk:
        return self.__getitem__(name)
    
    # ! IO Methods
    def readable(self) -> bool:
        return hasitem(self.__mode, 'r+')
    
    def seekable(self) -> bool:
        return self.__io.seekable()
    
    def writable(self) -> bool:
        return hasitem(self.__mode, 'w+')
    
    def flush(self) -> None:
        assert not self.__io.closed
        if not self.writable():
            raise IONotWritableError()
        self.__ioflush()
    
    def close(self) -> None:
        if self.writable():
            self.flush()
        for chunk in self.__chunks:
            if not chunk.closed:
                chunk.close()
        if not self.__io.closed:
            self.__io.close()


from typing import Generic
from .types import AT, KT, ErrorTextReturnType
from .units import CHUNKFILE_MODES

# ! Base Error Class
class Error(Exception, Generic[AT, KT]):
    """The base class of the error."""
    def __init__(self, *args: AT, **kwargs: KT) -> None:
        """The base class of the error."""
        super().__init__()
        self.args = tuple(" ".join([arg for arg in self.__text__(*args, **kwargs) if arg is not None]))
    
    def __text__(self, *args: AT, **kwargs: KT) -> ErrorTextReturnType:
        """Error text generator.
        
        Returns:
            ErrorTextReturnType: A generator with strings.
        """
        yield

# ! Chunkfile Errors
class IsNotChunkFileError(Error):
    def __text__(self, *args, **kwargs):
        yield "The IO you have opened is not a chunkfile."

class IsNotChunkFileModeError(Error[str, str]):
    def __text__(self, mode: str, *args, **kwargs):
        yield f"This mode value is not supported by the chunkfile: {repr(mode)}."
        yield f"Use anyone from this list: {', '.join([repr(cfmode) for cfmode in CHUNKFILE_MODES])}."

class ChunkIsDamagedError(Error):
    def __text__(self, *args, **kwargs):
        yield "The chunk is damaged."

# ! IO Errors
class IONotWritableError(Error):
    def __text__(self, *args, **kwargs):
        yield "This IO is open in a mode that does not provide for writing."


import os
import sys
from io import DEFAULT_BUFFER_SIZE
from pathlib import Path
# > Local Imports
from .units import CHUNKFILE_MODES
from .crc import CRC32
from .exceptions import IsNotChunkFileModeError

# ! IO Methods
def initfp(fp, mode):
    if isinstance(fp, (str, Path)):
        fp = os.path.abspath(str(fp))
        if mode in ['r', 'r+', '+r']:
            fpmode = 'rb+'
        elif mode in ['w', 'w+', '+w']:
            fpmode = 'wb+'
        else:
            raise IsNotChunkFileModeError(mode)
        return fp, open(fp, fpmode)
    else:
        assert not fp.closed
        assert fp.readable()
        assert fp.seekable()
        assert fp.writable()
        if hasattr(fp, "name"):
            if fp.name is not None:
                return fp.name, fp
        return None, fp

def iocopy(
    from_io,
    to_io,
    size,
    buffer_size=DEFAULT_BUFFER_SIZE,
    byteorder=sys.byteorder
):
    crc = CRC32(byteorder=byteorder)
    while size > 0:
        if size >= buffer_size:
            rsize, size = buffer_size, size - buffer_size
        else:
            rsize, size = size, 0
        d = from_io.read(rsize)
        crc.update(d)
        to_io.write(d)
    return crc

# ! Check Chunkfile Methods
def is_cfmode(value):
    return value in CHUNKFILE_MODES

# ! Check Methods
def hasitem(value, items):
    value = list(value)
    return max([i in value for i in items])

def hasitems(value, items):
    value = list(value)
    return min([i in value for i in items])

# ! Formatting Methods
def formater(**kwargs):
    return ", ".join([f"{key}={repr(value)}" for key, value in kwargs.items()])


import sys
from io import DEFAULT_BUFFER_SIZE
from typing import Callable, IO, Literal
from .chunkfile import ChunkFile, otempfile
from .types import FPType, CFMode
from .units import DEFAULT_CHUNK_FILE_SIGNATURE

# ! Main Open Method
def open(*args, **kwargs):
    return ChunkFile(*args, **kwargs)

"""Opening a chunkfile.
    
    Args:
        fp (FPType): The path or an already open IO.
        mode (CFMode, optional): The mode in which the file will be opened. Defaults to 'r'.
        buffer_size (int, optional): The size of the buffer. Defaults to DEFAULT_BUFFER_SIZE.
        chunk_file_signature (bytes, optional): The signature of the chunk file. Defaults to DEFAULT_CHUNK_FILE_SIGNATURE.
        chunk_name_size (int, optional): The length of the chunk name. Defaults to 4.
        chunk_length_size (int, optional): The number of bytes with the specified chunk size. Defaults to 4.
        ignore_crc (bool, optional): If True, then checksum verification does not occur. Defaults to False.
        openio (Callable[[], IO[bytes]], optional): The function of opening a temporary file. Defaults to otempfile.
    
    Returns:
        ChunkFile: A chunkfile.
    """


from types import TracebackType
from typing import IO, List, Iterable, Optional
# > Local Imports
from .types import CFMode
from .functional import is_cfmode, hasitem, formater

# ! Main Class
class Chunk:
    """The chunk."""
    def __init__(
        self,
        name: str,
        io: IO[bytes],
        mode: CFMode,
        *,
        chunk_name_size: int=4
    ) -> None:
        """Opening a chunk.
        
        Args:
            name (str): The name of the chunk.
            io (IO[bytes]): IO with chunk data.
            mode (CFMode): The mode of opening the chunk.
            chunk_name_size (int, optional): The length of the chunk name. Defaults to 4.
        """
        assert is_cfmode(mode)
        self.__io = io
        self.__mode = mode
        self.__name = name
        self.__cns = chunk_name_size
    
    # ! Propertyes
    @property
    def name(self) -> str:
        """The name of the chunk.
        
        Returns:
            str: The name of the chunk.
        """
        return self.__name
    
    @property
    def mode(self) -> CFMode:
        """The mode in which the chunk is open.
        
        Returns:
            CFMode: The mode in which the chunk is open.
        """
        return self.__mode
    
    @property
    def size(self) -> int:
        """The current size of the chunk data in bytes.
        
        Returns:
            int: The current size of the chunk data in bytes.
        """
        cp = self.__io.tell()
        self.__io.seek(0, 2)
        s = self.__io.tell()
        self.__io.seek(cp)
        return s
    
    @property
    def closed(self) -> bool:
        return self.__io.closed
    
    @property
    def _io(self) -> IO[bytes]:
        return self.__io
    
    # ! Magic Methods
    def __str__(self) -> str:
        return formater(name=self.__name, mode=self.__mode)
    
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__str__()})"
    
    def __enter__(self):
        return self
    
    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType]
    ) -> None:
        pass
    
    # ! Chunk Methods
    def rename(self, name: str) -> None:
        assert len(name) <= self.__cns
        self.__name = name.ljust(self.__cns)
    
    # ! Main IO Methods
    def readable(self) -> bool:
        return hasitem(self.__mode, ['r', '+'])
    
    def seekable(self) -> bool:
        return self.__io.seekable()
    
    def writable(self) -> bool:
        return hasitem(self.__mode, ['w', '+'])
    
    def read(self, __n: int=-1) -> bytes:
        assert self.readable()
        return self.__io.read(__n)
    
    def readline(self, __limit: int=-1) -> bytes:
        assert self.readable()
        return self.__io.readline(__limit)
    
    def readlines(self, __hint: int=-1) -> List[bytes]:
        assert self.readable()
        return self.__io.readlines(__hint)
    
    def tell(self) -> int:
        return self.__io.tell()
    
    def seek(self, __offset: int, __whence: int=0) -> int:
        assert self.seekable()
        return self.__io.seek(__offset, __whence)
    
    def write(self, __s: bytes) -> int:
        assert self.writable()
        return self.__io.write(__s)
    
    def writelines(self, __lines: Iterable[bytes]) -> None:
        assert self.writable()
        return self.__io.writelines(__lines)
    
    def flush(self) -> None:
        assert self.writable()
        return self.__io.flush()
    
    def truncate(self, __size: Optional[int]=None) -> int:
        assert self.seekable()
        assert self.writable()
        return self.__io.truncate(__size)
    
    def close(self) -> None:
        return self.__io.close()



from typing import List, Literal

DEFAULT_CHUNK_FILE_SIGNATURE: bytes = b"\x89CHKFL\r\n"
"""The default signature of the chunkfile."""
CHUNKFILE_MODES = ['r', 'r+', '+r', 'w', 'w+', '+w']
"""Modes for opening a chunkfile."""


import sys
from pathlib import Path
from io import BytesIO, BufferedRandom, DEFAULT_BUFFER_SIZE
# > Typing
from typing import Literal, Iterable, Tuple, Union, Optional, IO, overload
# > Local Imports
from .types import T, CFMode
from .crc import CRC32

# ! IO Methods
@overload
def initfp(fp: Union[str, Path], mode: CFMode) -> Tuple[str, BufferedRandom]: ...
@overload
def initfp(fp: BufferedRandom, mode: CFMode) -> Tuple[str, BufferedRandom]: ...
@overload
def initfp(fp: BytesIO, mode: CFMode) -> Tuple[None, BytesIO]: ...
@overload
def initfp(fp: IO[bytes], mode: CFMode) -> Tuple[Optional[str], IO[bytes]]: ...

def iocopy(
    from_io: IO[bytes],
    to_io: IO[bytes],
    size: int,
    buffer_size: int=DEFAULT_BUFFER_SIZE,
    byteorder=sys.byteorder
) -> CRC32: ...

# ! Has Methods
def hasitem(value: Iterable[T], items: Iterable[T]) -> bool: ...
def hasitems(value: Iterable[T], items: Iterable[T]) -> bool: ...

@overload
def is_cfmode(value: CFMode) -> Literal[True]: ...
@overload
def is_cfmode(value: str) -> Literal[False]: ...

# ! Formatting Methods
def formater(**kwargs: object) -> str: ...


from .crc import CRC32
from .chunk import Chunk
from .chunker import open
from .chunkfile import ChunkFile, otempfile, obytesio
from .types import CFMode, CFReadMode, CFReadUpdateMode, CFWriteMode, CFWriteUpdateMode
from .exceptions import IONotWritableError, IsNotChunkFileError, IsNotChunkFileModeError
from .units import DEFAULT_CHUNK_FILE_SIGNATURE, CHUNKFILE_MODES


__all__ = [
    'CRC32', 'Chunk', 'ChunkFile',
    'DEFAULT_CHUNK_FILE_SIGNATURE', 'CHUNKFILE_MODES',
    'CFMode', 'CFReadMode', 'CFReadUpdateMode', 'CFWriteMode', 'CFWriteUpdateMode',
    'IONotWritableError', 'IsNotChunkFileError', 'IsNotChunkFileModeError',
    'otempfile', 'obytesio', 'open'
]


import io
import sys
from typing import Callable, IO, Literal
from .chunkfile import ChunkFile, otempfile
from .types import FPType, CFMode
from .units import DEFAULT_CHUNK_FILE_SIGNATURE

# ! Main Open Method
def open(
    fp: FPType,
    mode: CFMode='r',
    *,
    buffer_size: int=io.DEFAULT_BUFFER_SIZE,
    chunk_file_signature: bytes=DEFAULT_CHUNK_FILE_SIGNATURE,
    chunk_name_size: int=4,
    chunk_length_size: int=4,
    openio: Callable[[], IO[bytes]]=otempfile,
    encoding: str='utf-8',
    errors: str='strict',
    byteorder: Literal['little', 'big']=sys.byteorder,
    ignore_crc: bool=False
) -> ChunkFile:
    """Opening a chunkfile.
    
    Args:
        fp (FPType): The path or an already open IO.
        mode (CFMode, optional): The mode in which the file will be opened. Defaults to 'r'.
        buffer_size (int, optional): The size of the buffer. Defaults to io.DEFAULT_BUFFER_SIZE.
        chunk_file_signature (bytes, optional): The signature of the chunk file. Defaults to DEFAULT_CHUNK_FILE_SIGNATURE.
        chunk_name_size (int, optional): The length of the chunk name. Defaults to 4.
        chunk_length_size (int, optional): The number of bytes with the specified chunk size. Defaults to 4.
        openio (Callable[[], IO[bytes]], optional): The function of opening a temporary file. Defaults to otempfile.
        encoding (str, optional): The name of the encoding used to decode the stream bytes into strings, and to encode strings into bytes. Defaults to 'utf-8'.
        errors (str, optional): The error setting of the decoder or encoder. Defaults to 'strict'.
        byteorder (Literal['little', 'big'], optional): An indicator of the native byte order. Defaults to sys.byteorder.
        ignore_crc (bool, optional): If True, then checksum verification does not occur. Defaults to False.
    
    Returns:
        ChunkFile: A chunkfile.
    """
    ...



from pathlib import Path
from io import BytesIO, BufferedRandom
from typing import (
    TypeVar,
    Generator,
    Union,
    Literal,
    Optional,
    IO,
    Any
)

# ! Type Vars
T = TypeVar('T')
AT = TypeVar('AT')
KT = TypeVar('KT')

# ! Error Types
ErrorTextReturnType = Generator[Optional[str], Any, None]

# ! IO Types
FPType = Union[str, Path, BytesIO, BufferedRandom, IO[bytes]]

# ! Chunk File Modes Types
CFReadMode = Literal['r']
CFWriteMode = Literal['w']
CFReadUpdateMode = Literal['r+', '+r']
CFWriteUpdateMode = Literal['w+', '+w']
CFMode = Union[CFReadMode, CFReadUpdateMode, CFWriteMode, CFWriteUpdateMode]


import sys
import zlib
from typing import Literal

# ! Main Class
class CRC32:
    """The checksum."""
    def __init__(
        self,
        start_value: int=0,
        byteorder: Literal['little', 'big']=sys.byteorder
    ) -> None:
        """The checksum.
        
        Args:
            start_value (int, optional): The starting value of the checksum. Defaults to 0.
        """
        self.__crc = start_value
        self.__byteorder = byteorder
    
    # ! Magic Methods
    def __str__(self) -> str:
        return str(self.__crc)
    
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__str__()})"
    
    def __int__(self) -> int:
        return self.__crc
    
    def __hex__(self) -> str:
        return hex(self.__crc)
    
    def __bytes__(self) -> bytes:
        return self.__crc.to_bytes(4, self.__byteorder)
    
    # ! Load Methods
    @staticmethod
    def from_bytes(
        data: bytes,
        byteorder: Literal['little', 'big']=sys.byteorder
    ):
        return CRC32(int.from_bytes(data, byteorder), byteorder)
    
    # ! Main Methods
    def update(self, data: bytes) -> None:
        """Updating the checksum.
        
        Args:
            data (bytes): A byte string for calculating the checksum.
        """
        self.__crc = zlib.crc32(data, self.__crc)
